local playerClass = select(2,UnitClass("player"))
if playerClass ~= "DRUID" and playerClass ~="MAGE" and playerClass ~="PALADIN" and playerClass ~="PRIEST" and playerClass ~="SHAMAN" and playerClass ~="WARLOCK" and playerClass ~= "ROGUE" then return end
local playerHealer, playerCaster, playerMelee, playerHybrid
if playerClass == "PRIEST" or playerClass == "SHAMAN" or playerClass == "PALADIN" or playerClass == "DRUID" then playerHealer = true end
if playerClass == "MAGE" or playerClass == "PRIEST" or playerClass == "WARLOCK" then playerCaster = true end
if playerClass == "DRUID" or playerClass == "PALADIN" or playerClass == "SHAMAN" then playerHybrid = true end
if playerClass == "ROGUE" or playerClass == "WARRIOR" or playerClass == "HUNTER" then playerMelee = true end

local L = AceLibrary("AceLocale-2.2"):new("DrDamage")
local BS
if GetLocale() ~= "enUS" and GetLocale() ~= "enGB" then BS = AceLibrary("Babble-Spell-2.2") end
local SEA = AceLibrary("SpecialEvents-Aura-2.0")
local GT = AceLibrary("Gratuity-2.0")
local Deformat = AceLibrary("Deformat-2.0")

--General
local settings
local type = type
local pairs = pairs
local ipairs = ipairs
local tonumber = tonumber
local math_abs = math.abs
local math_floor = math.floor
local math_ceil = math.ceil
local string_match = string.match
local select = select
local GetSpellName = GetSpellName
local UnitLevel = UnitLevel
local BOOKTYPE_SPELL = BOOKTYPE_SPELL
--General/caster
local UnitMana = UnitMana
local Mana_Cost = MANA_COST
--General/melee

--Core
local _G = getfenv(0)
local BOOKTYPE_SPELL = BOOKTYPE_SPELL
local GetSpellName = GetSpellName
local GetActionInfo = GetActionInfo
local HasAction = HasAction
local IsEquippedItem = IsEquippedItem
local DrD_Font = GameFontNormal:GetFont()
local playerCompatible
local updatingSpell, updateSetItems
DrDamage.visualChange = true

local function DrD_ClearTable( table )
	for k in pairs( table ) do
		table[k] = nil
	end
end

local function DrD_Round(x, y)
	return math_floor( x * 10 ^ y + 0.5 ) / 10 ^ y
end

local function DrD_MatchData( data, ... )
	if not data or not ... then
		return false
	end

	if type( data ) == "table" then
		for _, dataName in ipairs( data ) do
			for i = 1, select('#', ...) do
				if dataName == select(i, ...) then
					return true
				end
			end
		end
	else	
		for i = 1, select('#', ...) do
			if data == select(i, ...) then
				return true
			end
		end
	end
	
	return false
end

local function DrD_GetTextFrame( hide, ... )
	for i = 1, select('#', ...) do
		local frame = select(i, ...)
		if frame and frame:GetName() == "DrDamageText" and frame.text then
			if hide then frame.text:Hide() end
			return frame
		end
	end
end

function DrDamage:OnInitialize()
	if IsAddOnLoaded( "FlexBar2" ) then
		FlexBar2:RegisterTextSub("drd", "DrDamage_Update",
		function( Button, Frame, updateSpells )
			if not playerCompatible then return false end
			if Button:GetEffectiveAttribute('type') == 'action' then
				local id = Button:GetEffectiveAttribute( "action" )
				
				if id then
					--[[
					if updateSpells then

						local gtype, pid = GetActionInfo( id )

						if gtype == "spell" then --if gtype =="macro"?
							local spellName = GetSpellName( pid, BOOKTYPE_SPELL );	

							if DrD_MatchData( updateSpells, spellName ) then
								return self:CheckAction( id )
							end
						end
					else
					--]]
						return self:CheckAction( id )
					--end
				end
				return false
			end
		end)
	end
	if IsAddOnLoaded( "InfiniBar" ) then
		InfiniBar:RegisterTextSub("drd", "DrDamage_Update",
		function( bar, btn, options, updateSpells )
			if not playerCompatible then return false end
			local spellName, spellRank = bar:GetSpellNameAndRank( btn )
			if spellName and spellRank then
				if not updateSpells or DrD_MatchData( updateSpells, spellName ) then
					return self:CheckAction( nil, nil, nil, nil, spellName, spellRank )
				end
			end
			return false
		end)
	end
end

DrDamage:RegisterDB("DrDamageDB");
DrDamage:RegisterDefaults("profile", DrDamage.defaults)
DrDamage.name = "DrDamage"
DrDamage.title = "Dr. Damage"
DrDamage.hasNoColor = true
DrDamage.hasNoText = true
DrDamage.cannotDetachTooltip = true
DrDamage.hideWithoutStandby = true
DrDamage.independentProfile = true
DrDamage.blizzardTooltip = true
if playerMelee then
	DrDamage.hasIcon = "Interface\\Icons\\Ability_DualWield"
else
	DrDamage.hasIcon = "Interface\\Icons\\Spell_Holy_SearingLightPriest"
end

function DrDamage:OnEnable()
	if self.PlayerData then
		self.ClassSpecials = {}
		self.Calculation = {}
		self.SetBonuses = {}	
		
		if DrDamage.Caster_OnEnable then
			DrDamage:Caster_OnEnable()
		end
		if DrDamage.Melee_OnEnable then
			DrDamage:Melee_OnEnable()
		end
		
		self:PlayerData()
		self.PlayerData = nil
		playerCompatible = true
	elseif not DrDamage.spellInfo and not DrDamage.talentInfo then
		return
	end
	
	self:RegisterChatCommand({ "/drdmg", "/drdamage" }, self.options)
	self.OnMenuRequest = self.options	
	
	self.talents = {}	
	settings = self.db.profile
	
	self:DmgUpdateTalents()
	
	self:Hook(GameTooltip, "SetAction", true)
	self:SecureHook(GameTooltip, "SetSpell")
    	self:RegisterEvent("CHARACTER_POINTS_CHANGED")
    	self:RegisterEvent("PLAYER_ALIVE")
    	self:RegisterEvent("AceDB20_ResetDB")
    			
	if settings.DisplayType ~= "Off" then
    		self:RegisterEvent("ACTIONBAR_PAGE_CHANGED")
    		self:RegisterEvent("ACTIONBAR_HIDEGRID")
    		self:RegisterEvent("UPDATE_SHAPESHIFT_FORM")
    		self:RegisterEvent("PLAYER_TARGET_CHANGED")
    		self:RegisterEvent("UPDATE_MACROS")
    		self:RegisterEvent("UNIT_INVENTORY_CHANGED")
    		
		if self.PlayerAura then
    			self:RegisterEvent( "SpecialEvents_PlayerBuffGained", "PlayerAuraUpdate" )
    			self:RegisterEvent( "SpecialEvents_PlayerBuffCountChanged", "PlayerAuraUpdate" )
    			self:RegisterEvent( "SpecialEvents_PlayerBuffLost", "PlayerAuraUpdate" )
    			self:RegisterEvent( "SpecialEvents_PlayerDebuffGained", "PlayerAuraUpdate" )
    			self:RegisterEvent( "SpecialEvents_PlayerDebuffCountChanged", "PlayerAuraUpdate" )
    			self:RegisterEvent( "SpecialEvents_PlayerDebuffLost", "PlayerAuraUpdate" )
    		end
    		if playerHealer and self.HealingBuffs then
    			self:RegisterEvent( "SpecialEvents_UnitBuffGained", "TargetBuffUpdate" )
    			self:RegisterEvent( "SpecialEvents_UnitBuffCountChanged", "TargetBuffUpdate" )
    			self:RegisterEvent( "SpecialEvents_UnitBuffLost", "TargetBuffUpdate" )
    		end
    		if self.Debuffs then
    			self:RegisterEvent( "SpecialEvents_UnitDebuffGained", "TargetDebuffUpdate" )
    			self:RegisterEvent( "SpecialEvents_UnitDebuffCountChanged", "TargetDebuffUpdate" )
    			self:RegisterEvent( "SpecialEvents_UnitDebuffLost", "TargetDebuffUpdate" )
    		end
  	end
end

function DrDamage:OnDisable()
	self.talents = nil
	
	if playerCompatible then
		local tempType = settings.DisplayType
		settings.DisplayType = "Off"
		self:UpdateAB()
		settings.DisplayType = tempType
	end	
end

function DrDamage:OnProfileEnable()
	DrDamage:Caster_OnProfileEnable()
	settings = self.db.profile
end

function DrDamage:CHARACTER_POINTS_CHANGED()
    	self:DmgUpdateTalents()
end

function DrDamage:PLAYER_ALIVE()
    	self:DmgUpdateTalents()
    	self:UnregisterEvent("PLAYER_ALIVE")
end

function DrDamage:ACTIONBAR_PAGE_CHANGED()
	if not self:IsEventScheduled("UpdatingAB") or updatingSpell then
		self:ScheduleEvent("UpdatingAB", self.UpdateAB, 0.25, self)
	end
	self.visualChange = true
end

function DrDamage:ACTIONBAR_HIDEGRID()
	if (not self:IsEventScheduled("UpdatingAB") or updatingSpell) and ( GetCursorInfo() == "spell" or GetCursorInfo() == "macro" )  then
		self:ScheduleEvent("UpdatingAB", self.UpdateAB, 0.5, self)
	end
	self.visualChange = true
end

function DrDamage:UPDATE_SHAPESHIFT_FORM()
	if not self:IsEventScheduled("UpdatingAB") or updatingSpell then
		 self:ScheduleEvent("UpdatingAB", self.UpdateAB, 0.5, self)
	end
	self.visualChange = true
end

function DrDamage:PLAYER_TARGET_CHANGED()
	if not self:IsEventScheduled("UpdatingAB") or updatingSpell then
		self:ScheduleEvent("UpdatingAB", self.UpdateAB, 2, self)
	end
end

function DrDamage:UPDATE_MACROS()
	if not self:IsEventScheduled("UpdatingAB") or updatingSpell then
		self:ScheduleEvent("UpdatingAB", self.UpdateAB, 1, self)
	end
	self.visualChange = true
end

function DrDamage:UNIT_INVENTORY_CHANGED( unit )
	if unit == "player" then
		if DrDamage.Caster_InventoryChanged then
			if DrDamage:Caster_InventoryChanged() then
				updateSetItems = true
				self:UpdateAB()
				return
			end
		end
		if DrDamage.Melee_InventoryChanged then
			if DrDamage:Melee_InventoryChanged() then
				updateSetItems = true
				self:UpdateAB()
				return
			end
		end
	end
end

function DrDamage:PlayerAuraUpdate( buffName )
	if self:IsEventScheduled("UpdatingAB") and not updatingSpell then
		return
	end
	
	if playerCaster or playerHybrid then
		if self:Caster_CheckBaseStats() then
			self:UpdateAB()
			return
		end
	elseif playerMelee then
		if self:Melee_CheckBaseStats() then
			self:UpdateAB()
			return
		end
	end
	
	if BS and BS:HasReverseTranslation( buffName ) then
		buffName = BS:GetReverseTranslation( buffName )
	end
				
	if self.PlayerAura[buffName] then
		local updateSpell = self.PlayerAura[buffName].Spell
		
		if updatingSpell and updatingSpell == updateSpell or updateSpell then
			self:ScheduleEvent("UpdatingAB", self.UpdateAB, 0.5, self, updateSpell)
			updatingSpell = updateSpell
		else
			self:ScheduleEvent("UpdatingAB", self.UpdateAB, 0.5, self)
		end
	end
end

function DrDamage:TargetBuffUpdate( unitID, buffName )	
	if unitID == "target" then
		if self:IsEventScheduled("UpdatingAB") and not updatingSpell then
			return
		end
			
		if BS and BS:HasReverseTranslation( buffName ) then
			buffName = BS:GetReverseTranslation( buffName )
		end		
		
		if self.HealingBuffs[buffName] then
			local buff = self.HealingBuffs[buffName]

			if buff.Class and DrD_MatchData( buff.Class, playerClass ) or not buff.Class then
				local updateSpell = buff.Spell
				if updatingSpell and updatingSpell == updateSpell or updateSpell then
					self:ScheduleEvent("UpdatingAB", self.UpdateAB, 0.5, self, updateSpell )
					updatingSpell = updateSpell				
				else
					self:ScheduleEvent("UpdatingAB", self.UpdateAB, 1.0, self)
				end
			end
		end
	end
end

function DrDamage:TargetDebuffUpdate( unitID, buffName )	
	if unitID == "target" then
		if self:IsEventScheduled("UpdatingAB") and not updatingSpell then
			return
		end
			
		if BS and BS:HasReverseTranslation( buffName ) then
			buffName = BS:GetReverseTranslation( buffName )
		end		
		
		if self.Debuffs[buffName] then
			local debuff = self.Debuffs[buffName]
		
			if debuff.Class and DrD_MatchData( debuff.Class, playerClass ) or playerHealer and debuff.School == "Healing" or not debuff.Class and not debuff.School then
				local updateSpell = debuff.Spell
				if updatingSpell and updatingSpell == updateSpell or updateSpell then
					self:ScheduleEvent("UpdatingAB", self.UpdateAB, 0.75, self, updateSpell )
					updatingSpell = updateSpell				
				else
					self:ScheduleEvent("UpdatingAB", self.UpdateAB, 1.5, self)
				end
			end
		end
	end
end

local lastState = GetTime()
function DrDamage:MODIFIER_STATE_CHANGED( state )
	if state == "ALT" and settings.UpdateAlt or state == "CTRL" and settings.UpdateCtrl or state == "SHIFT" and settings.UpdateShift then
		if GetTime() - lastState < 0.3 then
			if self:IsEventScheduled("Modifier_UpdatingAB") then
				self:CancelScheduledEvent( "Modifier_UpdatingAB" )
				self.visualChange = false
			end
		else
			if not self:IsEventScheduled("Modifier_UpdatingAB") then
				self:ScheduleEvent("Modifier_UpdatingAB", self.UpdateAB, 0.3, self)
			end
			self.visualChange = true
		end
		lastState = GetTime()
	end
end

function DrDamage:AceDB20_ResetDB( name, dab )
	if dab == self.db.name then
		if not self:IsEventScheduled("UpdatingAB") then
			self:ScheduleEvent("UpdatingAB", self.UpdateAB, 1, self)
		end
		self.visualChange = true
	end
end

function DrDamage:OnTooltipUpdate()
	GameTooltip:AddLine( "          Dr. Damage" )
	GameTooltip:AddLine( "Hint: Right-Click for options", 0, 1, 0, true )
end

function DrDamage:DmgUpdateTalents()
	DrD_ClearTable( self.talents )

	for t = 1, GetNumTalentTabs() do	
		for i = 1, GetNumTalents(t) do
			local nameTalent, _, _, _, currRank = GetTalentInfo( t, i )

			if BS and BS:HasReverseTranslation( nameTalent ) then
				nameTalent = BS:GetReverseTranslation( nameTalent )
			end

			if currRank ~= 0 then
				if DrDamage.talentInfo[nameTalent] then
					if DrDamage.talentInfo[nameTalent].Locale then
						if DrD_MatchData( DrDamage.talentInfo[nameTalent].Locale, GetLocale() ) then
							self.talents[nameTalent] = currRank
						end
					else
						self.talents[nameTalent] = currRank
					end
					--self:Print("Talent name: " .. nameTalent .. " Points: " .. currRank )
				end
			end	    	    
		end
	end
	if not self:IsEventScheduled("UpdatingAB") then
		self:ScheduleEvent("UpdatingAB", self.UpdateAB, 1.5, self)
	end
end

local SetItems = {}
function DrDamage:GetSetAmount(set)
	if not updateSetItems and SetItems[set] then
		return SetItems[set]
	end
	
	if updateSetItems then
		DrD_ClearTable( SetItems )
		updateSetItems = false
	end
	
	local amount = 0
	local setData = self.SetBonuses[set]
	if setData then
		for _, itemID in ipairs(setData) do
			if IsEquippedItem(itemID) then
				amount = amount + 1
			end
		end
	end
	SetItems[set] = amount
	return amount
end

function DrDamage:SetSpell( frame, slot )
	if not settings.Display then
		do return end
	end
		
	GT:SetSpell( slot, BOOKTYPE_SPELL )
	local spellName, spellRank = GetSpellName( slot, BOOKTYPE_SPELL );
	
	if spellRank then 
		spellRank = tonumber(string_match(spellRank, "%d+"))
	end
	
	if not spellRank then
		spellRank = "None"
	end
	
	if playerCaster or playerHybrid then
		self:DoTooltip( frame, spellName, spellRank )
	end
end

function DrDamage:SetAction( frame, slot )
	self.hooks[frame].SetAction(frame, slot)
	
	if not settings.Display then
		do return end
	end

	GT:SetAction(slot)	
	local gtype = GetActionInfo(slot)
	local spellName, spellRank = frame:GetSpell()

	if gtype == "spell" or gtype == "macro" then
		if spellName then 
			if spellRank then 
				spellRank = tonumber(string_match(spellRank, "%d+"))
			end
			if not spellRank then
				spellRank = "None"
			end
			if playerCaster or playerHybrid then
				self:DoTooltip( frame, spellName, spellRank )
			end
		end
	end
end
	
local ABstop, ABdefault, ABtable, ABgetID, ABglobalbutton, CTBar, ABdisable
local defaultBars = { "MultiBarBottomLeftButton", "MultiBarBottomRightButton","MultiBarRightButton", "MultiBarLeftButton" }

local function DrD_DetermineAB()
	if IsAddOnLoaded("Bartender3") then
		ABglobalbutton = "BT3Button"
		ABtable = Bartender3.actionbuttons
		ABgetID = function( button, index, ABtable )
					if ABtable[index] and ABtable[index].state == "used" then
						return ABtable[index].object:PagedID()
					end
		end	
	elseif IsAddOnLoaded( "Bongos2" ) then
		ABglobalbutton = "BongosActionButton"
		ABgetID = function( button ) return SecureButton_GetModifiedAttribute(button,"action",SecureStateChild_GetEffectiveButton(button)) end
	elseif IsAddOnLoaded( "Bongos" ) then
		ABglobalbutton = "BActionButton"
		ABgetID = function( button ) return SecureButton_GetModifiedAttribute(button,"action",SecureStateChild_GetEffectiveButton(button)) end
	elseif IsAddOnLoaded( "CogsBar" ) then
		ABglobalbutton = "CogsBarButton"
		ABgetID = function( button ) return SecureButton_GetModifiedAttribute(button,"action",SecureStateChild_GetEffectiveButton(button)) end
	elseif IsAddOnLoaded("TrinityBars") then
		ABglobalbutton = "TrinityActionButton"
		ABgetID = function( button ) return button:GetID() end
	elseif IsAddOnLoaded( "InfiniBar" ) or IsAddOnLoaded( "FlexBar2" ) then
		ABstop = true
	elseif IsAddOnLoaded("DiscordActionBars") then
		ABglobalbutton = "DAB_ActionButton_"
		ABgetID = function( button ) return button:GetActionID() end
	elseif IsAddOnLoaded("Nurfed") then
		ABglobalbutton = "Nurfed_Button"
		ABgetID = function( button )
			if button.type == "spell" and button.spell then
				local pos = string.find( button.spell, "%(" )
				if pos then pos = pos - 1
				else pos = string.len( button.spell ) end
				return nil, string.sub( button.spell, 1, pos ), string_match(button.spell,"%d+")
			elseif button.type == "macro" and button.spell then
				local action = GetActionFromMacroText(GetMacroBody(button.spell))
				local pid
				if action and not GetItemInfo(action) then pid = Nurfed:getspell(action) end
				if pid then return nil, GetSpellName(pid, BOOKTYPE_SPELL), pid end			
			end
		end		
	elseif IsAddOnLoaded("CT_BarMod") then
		CTBar = true
		ABtable = CT_BarMod.actionButtonList
		ABgetID = function( button, index, ABtable )
			if ABtable[index] and ABtable[index].hasAction then return index end
		end		
	else
		ABdefault = true
	end
	
	if settings.UpdateAlt or settings.UpdateCtrl or settings.UpdateShift then
		DrDamage:RegisterEvent( "MODIFIER_STATE_CHANGED" )
	end
end

local DrD_ProcessButton
function DrDamage:UpdateAB(updateSpell, manaUpdate)

	self:TriggerEvent("DrDamage_Update", updateSpell )

	--[[	Used for debugging updates.
	if updateSpell then
		self:Print( "Update AB: Spell" )
		
		if type( updateSpell ) == "table" then
			for _, gar in ipairs( updateSpell ) do
				self:Print( "Updating: " .. gar )
			end
		else
			self:Print( "Updating: " .. updateSpell )
		end
	else
		self:Print( "Update AB: All" )
	end
	--]]

	if DrD_DetermineAB then 
		DrD_DetermineAB()
		DrD_DetermineAB = nil
	end
	if not ABdefault then
		if settings.BlizzardAB then 
			ABdefault = true 
		elseif ABstop then
			updatingSpell = nil
			return
		end
	end
	
	if ABglobalbutton then
		for i = 1, 120 do	
			DrD_ProcessButton(_G[ABglobalbutton..i], i, updateSpell, manaUpdate)
		end
	elseif CTBar then
		for i, list in pairs(ABtable) do
			DrD_ProcessButton(list.button, i, updateSpell, manaUpdate)
		end	
	end
	
	if ABdefault or CTBar then
		if ABdefault then
			if not settings.BlizzardAB and (CTBar or ABglobalbutton) then
				ABdisable = true
				ABdefault = false
			end
			for bar = 1, 4 do
				for i = 1, 12 do
					DrD_ProcessButton(_G[defaultBars[bar]..i], nil, updateSpell, manaUpdate, ABdisable)
				end
			end			
		end
		if _G["BonusActionBarFrame"]:IsVisible() then
			for i = 1, 12 do
				DrD_ProcessButton(_G["BonusActionButton"..i], nil, updateSpell, manaUpdate, ABdisable)
			end
		else
			for i = 1, 12 do		
				DrD_ProcessButton(_G["ActionButton"..i], nil, updateSpell, manaUpdate, ABdisable)
			end
		end
	end
					
	updatingSpell = nil
	self.visualChange = nil
	ABdisable = false
end

DrD_ProcessButton = function( button, index, updateSpell, manaUpdate, ABdisable )
	if not button then do return end end

	local textFrame = DrD_GetTextFrame(not updateSpell,button:GetChildren())
	
	if settings.DisplayType == "Off" or ABdisable then
		do return end
	end

	if button:IsVisible() then
		local id, spellName, spellRank, pid
		
		if not index then 
			id = ActionButton_GetPagedID(button)
		elseif ABgetID then
			id, spellName, spellRank, pid = ABgetID(button, index, ABtable) 
		end

		if id then
			if HasAction(id) then
				DrDamage:CheckAction(id, button, manaUpdate, updateSpell, nil, nil, textFrame)
			end
		elseif spellName and spellRank then
			DrDamage:CheckAction(nil, button, manaUpdate, updateSpell, spellName, spellRank, textFrame, pid)
		end
	end
end

local displayTypeTable = { ["Avg"] = 2, ["DPS"] = 3, ["DPSC"] = 4, ["AvgHit"] = 5, ["Min"] = 6, ["MaxHit"] = 7, ["AvgHitTotal"] = 8, ["Max"] = 9, ["DPM"] = 10, ["ManaCost"] = 11, ["TrueManaCost"] = 12, ["MPS"] = 13 }

function DrDamage:CheckAction(id, button, manaUpdate, updateSpell, spellName, spellRank, textFrame, pid)
	local gtype
	
	if settings.DisplayType == "Off" then return "" end
	
	if id then
		gtype, pid = GetActionInfo(id)

		if gtype == "spell" then
			spellName, spellRank = GetSpellName(pid, BOOKTYPE_SPELL)
		elseif gtype == "macro" then
			GT:SetAction(id)
			spellName = GT:GetLine(1, false)
			spellRank = GT:GetLine(1, true)

			if not spellName and not spellRank and button then
				local macroText = _G[button:GetName().."Name"]
				if macroText then macroText:Show() end			
			end
		end
	end
	if spellName then
		if updateSpell then
			if not DrD_MatchData(updateSpell, spellName) then return end
			--self:Print( "Spell matched." )
		end	
	
		if self.ClassSpecials[spellName] then
			return self:SpecialSlot(self.ClassSpecials[spellName], spellRank, button, nil, textFrame)
		end
	
		if spellRank and DrDamage.spellInfo[spellName] then
			local baseSpell = DrDamage.spellInfo[spellName][0]
			local tableSpell = DrDamage.spellInfo[spellName][tonumber(string_match(spellRank,"%d+")) or "None"]

			if tableSpell then
				if gtype == "macro" and button then
					local macroText = _G[button:GetName().."Name"]
					if macroText then macroText:Hide() end
					if CTBar then if button.name then button.name:SetText("") end end
				end
				
				local text, healingSpell
				
				if playerCaster or playerHybrid then
					healingSpell = baseSpell.Healing or DrD_MatchData(baseSpell.School, "Healing")
					local selector = displayTypeTable[settings.DisplayType]

					if id or pid and not baseSpell.NoDPM then
						local casts
						if settings.CastsLeft and healingSpell or settings.CastsLeftDmg and not healingSpell then
							casts = true
						end
						if selector > 9 or casts then
							if id then GT:SetAction(id)
							else GT:SetSpell(pid, BOOKTYPE_SPELL) end
							if casts then
								if GT:GetLine(2) and Mana_Cost then
									manaCost = Deformat(GT:GetLine(2), Mana_Cost)
								end
								if manaCost then
									manaCost = tonumber(manaCost)
									if manaCost > 0 then
										text = math.floor( UnitMana("player") / manaCost + 0.5 )
									else
										text = "\226\136\158"
									end
								end
							else
								local values = select(10, DrDamage:RawNumbers( tableSpell, spellName, false, true ))
								text = values[settings.DisplayType]
							end
						end
					end

					if not text then
						if manaUpdate then
							return
						else
							if baseSpell.NoDPS and (selector == 3 or selector == 4) then selector = 2 end
							if selector > 9 then selector = 2 end
							text = math.floor((select(selector, self:RawNumbers(tableSpell, spellName)))+0.5)
						end
					else
						if type(text) == "number" then
							text = DrD_Round(text,1)
						end								
					end
				elseif playerMelee then
					text = DrDamage:MeleeCalc( tableSpell, spellName, false )
				end

				if button then
					self:CreateABtext(button, text, healingSpell, textFrame)
				else
					return text
				end
			end				
		end
	end
end

function DrDamage:CreateABtext( button, text, healingSpell, frame, r, g, b )

	if not frame then
		frame = DrD_GetTextFrame(false,button:GetChildren())
	end
	
	if frame and self.visualChange or not frame then
		if not r then
			if healingSpell then
				local color = settings.FontColorHeal
				r,g,b = color.r, color.g, color.b
			else
				local color = settings.FontColorDmg
				r,g,b = color.r, color.g, color.b
			end
		end
	end		
	
	if frame then
		if self.visualChange then
			frame.text:SetFont(DrD_Font, settings.FontSize, settings.FontEffect)
			frame.text:SetPoint("BOTTOMLEFT", frame ,"BOTTOMLEFT", -10 + settings.FontXPosition, settings.FontYPosition + 5)
			frame.text:SetPoint("BOTTOMRIGHT", frame ,"BOTTOMRIGHT", 10 + settings.FontXPosition, settings.FontYPosition + 5)
			frame.text:SetTextColor(r,g,b)	
		end
		frame.text:SetText(text)
		frame.text:Show()
	else
		local dpsText = CreateFrame("Frame", "DrDamageText", button)
		dpsText:SetAllPoints(button)
		dpsText:SetFrameLevel(dpsText:GetFrameLevel() + 1)
		dpsText.text = dpsText:CreateFontString( nil, "OVERLAY")
		dpsText.text:SetPoint("BOTTOMLEFT", dpsText ,"BOTTOMLEFT", -10 + settings.FontXPosition, settings.FontYPosition + 5)
		dpsText.text:SetPoint("BOTTOMRIGHT", dpsText ,"BOTTOMRIGHT", 10 + settings.FontXPosition, settings.FontYPosition + 5)
		dpsText.text:SetFont(DrD_Font, settings.FontSize, settings.FontEffect)
		dpsText.text:SetJustifyH("CENTER")
		dpsText.text:SetTextColor(r,g,b)
		dpsText.text:SetText(text)
		dpsText.text:Show()
		
		--Fix to properly inherit parent alpha
		if button:GetAlpha() == 0 then
			button:SetAlpha(0.01)
			button:SetAlpha(0)
		end
	end
end

function DrDamage:SpecialSlot( func, rank, button, healingSpell, textFrame )
	if rank then rank = tonumber(string_match(rank,"%d+")) end
	local text, r, g, b = func(rank)
	if type(text) == "number" then text = math_floor(text + 0.5) end
	if button and text then self:CreateABtext(button, text, healingSpell, textFrame, r, g, b)
	elseif text then return text end
end

--Credits to the author of RatingBuster (Whitetooth) for the formula!
local ratingTypes = { ["Hit"] = 8, ["Crit"] = 14, ["MeleeHit"] = 10 }

function DrDamage:GetRating( rType, convertR, full )
	local playerLevel = UnitLevel("player")
	local base = ratingTypes[rType]
	local rating

	if playerLevel < 10 then
		playerLevel = 10
	end
	if playerLevel <= 60 then
		rating = base / (( playerLevel - 8 ) / 52)
	elseif playerLevel <= 70 then
		rating = base / (131/41 - 3/82 * playerLevel)
	end

	if full then
		if convertR then
			return (convertR / rating)
		else
			return rating
		end
	else
		if convertR then
			return DrD_Round(convertR / rating, 2)
		else
			return DrD_Round(rating, 2)
		end
	end
end

function DrDamage:TotalAP()
	local baseAP, posBuff, negBuff = UnitAttackPower("player")
	return baseAP + posBuff + negBuff
end

function DrDamage:GetSpellInfo( spellName, spellRank )
	if spellName and spellRank then
		if DrDamage.spellInfo and DrDamage.spellInfo[spellName] then
			return DrDamage.spellInfo[spellName][spellRank] or DrDamage.spellInfo[spellName]["None"]
		end
	end
end